# imports

import pyomo.environ as pyo
from .other import price_other_block, price_other_no_block
from .delta import price_delta_block, price_delta_no_block
from .lambda_ import price_lambda_block, price_lambda_no_block

# *****************************************************************************
# *****************************************************************************

NODE_PRICE_LAMBDA = 'lambda'
NODE_PRICE_DELTA = 'delta'
NODE_PRICE_OTHER = None
NODE_PRICES = (
    NODE_PRICE_LAMBDA,
    NODE_PRICE_DELTA,
    NODE_PRICE_OTHER
    )

def add_price_functions(
    model: pyo.AbstractModel,
    use_blocks: bool = True,
    node_price_model = NODE_PRICE_OTHER,
    **kwargs
):  
    
    # *************************************************************************
    # *************************************************************************
    
    if use_blocks:
        # with blocks
        if node_price_model == NODE_PRICE_LAMBDA:
            price_lambda_block(model, **kwargs)
        elif node_price_model == NODE_PRICE_DELTA:
            price_delta_block(model, **kwargs)
        else:
            price_other_block(model, **kwargs)
    else:
        # without blocks
        if node_price_model == NODE_PRICE_LAMBDA:
            price_lambda_no_block(model, **kwargs)
        elif node_price_model == NODE_PRICE_DELTA:
            price_delta_no_block(model, **kwargs)
        else:
            price_other_no_block(model, **kwargs)
        
    # *************************************************************************
    # *************************************************************************

# *****************************************************************************
# *****************************************************************************